// -*- C++ -*- (c) 2016 Vladimír Štill <xstill@fi.muni.cz>

DIVINE_RELAX_WARNINGS
#include <llvm/Pass.h>
#include <llvm/IR/Module.h>
#include <llvm/IR/Instructions.h>
#include <llvm/IR/IRBuilder.h>
#include <llvm/Analysis/LoopInfo.h>
#include <llvm/IR/Dominators.h>
DIVINE_UNRELAX_WARNINGS

#include <brick-string>
#include <brick-llvm>
#include <string>
#include <iostream>

#include <lart/support/pass.h>
#include <lart/support/meta.h>
#include <lart/support/query.h>
#include <lart/support/util.h>
#include <lart/support/metadata.h>
#include <lart/divine/passes.h>
#include <lart/reduction/passes.h>

#include <divine/vm/divm.h>

namespace lart {
namespace divine {

struct CflInterrupt {

    static PassMeta meta() {
        return passMeta< CflInterrupt >( "CflInterrupt", "Instrument all control flow cycles "
                                         "with __vm_interrupt_cfl intrinsic." );
    }

    void annotateFn( llvm::Function &fn ) {

        for ( auto b : getBackEdges( fn ) ) {
            auto *src = b.from;
            auto idx = b.succIndex;
            auto *term = src->getTerminator();
            auto *dst = term->getSuccessor( idx );
            ASSERT_LEQ( 1, term->getNumSuccessors() );

            // Try to be smart and avoid invoking unnecessary interrupts which
            // are not on back edges.
            // However, for IndirectBr, there is no way to be smart as the
            // successors in the code are only hints for LLVM and real
            // successors are determined at runtime and therefore cannot be
            // rewritten -- so for IndirectBr we fallback to instrumenting
            // eagerly.
            if ( term->getNumSuccessors() == 1
                    || term->getOpcode() == llvm::Instruction::IndirectBr )
            {
                llvm::IRBuilder<>( term ).CreateCall( _cflInterrupt, { } );
            }
            else if ( dst->getUniquePredecessor() )
            {
                llvm::IRBuilder<>( dst->getFirstInsertionPt() ).CreateCall( _cflInterrupt, { } );
            }
            else
            {
                // neither of the blocks of the edge is used only by this path,
                // insert backedge block for savepc and fix branch and PHI nodes
                auto *back = llvm::BasicBlock::Create( fn.getParent()->getContext(), "backedge", &fn );
                llvm::IRBuilder<> irb( back );
                irb.CreateCall( _cflInterrupt, { } );
                irb.CreateBr( dst );

                // re-wire terminator and phi nodes
                term->setSuccessor( idx, back );
                for ( auto &inst : *dst ) {
                    auto *phi = llvm::dyn_cast< llvm::PHINode >( &inst );
                    if ( !phi )
                        break;
                    for ( unsigned i = 0, end = phi->getNumIncomingValues(); i < end; ++i ) {
                        if ( phi->getIncomingBlock( i ) == src )
                            phi->setIncomingBlock( i, back );
                    }
                }
            }
            ++_backedges;
        }
    }

    void run( llvm::Module &m ) {
        if ( !tagModuleWithMetadata( m, "lart.divine.interrupt.cfl" ) )
            return;

        auto *spcty = llvm::FunctionType::get( llvm::Type::getVoidTy( m.getContext() ), false );
        _cflInterrupt = llvm::cast< llvm::Function >( m.getOrInsertFunction( "__vm_interrupt_cfl", spcty ) );
        ASSERT( _cflInterrupt );
        _cflInterrupt->addFnAttr( llvm::Attribute::NoUnwind );

        std::set< llvm::Function * > skip;
        brick::llvm::enumerateFunctionsForAnno( "lart.interrupt.skipcfl", m,
                                                [&]( llvm::Function *f ) { skip.insert( f ); } );

        for ( auto &fn : m ) {
            if ( fn.empty() || skip.count( &fn ) )
                continue;
            annotateFn( fn );
        }
        // std::cout << "Found " << _backedges << " backedges" << std::endl;
    }

    llvm::Function *_cflInterrupt;
    long _backedges = 0;
};

struct MemInterrupt {

    static PassMeta meta() {
        return passMeta< MemInterrupt >( "MemInterrupt", "Instrument all memory accesses with "
                                         "__vm_interrupt_mem intrinsic." );
    }

    void annotateFn( llvm::Function &fn, llvm::DataLayout &dl, unsigned silentID ) {
        // avoid changing bb while we iterate over it
        for ( auto inst : query::query( fn ).flatten().map( query::refToPtr ).freeze() ) {
            auto op = inst->getOpcode();
            if ( ( op == llvm::Instruction::Load || op == llvm::Instruction::Store
                    || op == llvm::Instruction::AtomicRMW
                    || op == llvm::Instruction::AtomicCmpXchg )
                  && !reduction::isSilent( *inst, silentID ) )
            {
                auto *type = _memInterrupt->getFunctionType();
                auto point = llvm::BasicBlock::iterator( inst );
                llvm::IRBuilder<> irb{ point };
                auto *origPtr = getPointerOperand( inst );
                auto *origT = llvm::cast< llvm::PointerType >( origPtr->getType() )
                                  ->getElementType();
                auto *ptr = irb.CreateBitCast( origPtr, type->getParamType( 0 ) );
                auto *si = irb.getInt32( std::max( uint64_t( 1 ), dl.getTypeSizeInBits( origT ) / 8 ) );
                point ++;
                int intr_type;
                switch ( op )
                {
                    case llvm::Instruction::Load: intr_type = _VM_MAT_Load; break;
                    case llvm::Instruction::Store: intr_type = _VM_MAT_Store; break;
                    default: intr_type = _VM_MAT_Both; break;
                }
                irb.SetInsertPoint( point );
                irb.CreateCall( _memInterrupt, { ptr, si, irb.getInt32( intr_type ) } );
                ++_mem;
            }
        }
    }

    void run( llvm::Module &m ) {
        if ( !tagModuleWithMetadata( m, "lart.divine.interrupt.mem" ) )
            return;

        llvm::DataLayout dl( &m );
        auto &ctx = m.getContext();
        auto *i32t = llvm::Type::getInt32Ty( ctx );
        auto *ty = llvm::FunctionType::get( llvm::Type::getVoidTy( ctx ),
                            { llvm::Type::getInt8PtrTy( ctx ), i32t, i32t }, false );
        _memInterrupt = llvm::cast< llvm::Function >( m.getOrInsertFunction( "__vm_interrupt_mem", ty ) );
        ASSERT( _memInterrupt );
        _memInterrupt->addFnAttr( llvm::Attribute::NoUnwind );

        auto silentID = m.getMDKindID( reduction::silentTag );

        for ( auto &fn : m ) {
            if ( fn.empty() )
                continue;
            annotateFn( fn, dl, silentID );
        }
        // std::cout << "Found " << _mem << " memory accesses" << std::endl;
    }

    llvm::Function *_memInterrupt;
    long _mem = 0;
};

PassMeta interruptPass()
{
    return compositePassMeta< CflInterrupt, MemInterrupt >(
        "interrupt",
        "Instrument bitcode for DIVINE: add __vm_interrupt_mem and "
        "__vm_interrupt_cfl calls to appropriate locations." );
}

PassMeta cflInterruptPass()
{
    return passMeta< CflInterrupt >(
        "interrupt-cfl", "add __vm_interrupt_cfl calls to appropriate locations." );
}

}
}

